//#define TGA_DEBUG



/*
       0  -  No image data included.
       1  -  Uncompressed, color-mapped images.
       2  -  Uncompressed, RGB images.
       3  -  Uncompressed, black and white images.
       9  -  Runlength encoded color-mapped images.
      10  -  Runlength encoded RGB images.
      11  -  Compressed, black and white images.
      32  -  Compressed color-mapped data, using Huffman, Delta, and
                    runlength encoding.
      33  -  Compressed color-mapped data, using Huffman, Delta, and
                    runlength encoding.  4-pass quadtree-type process.

This document will describe only four formats: 1, 2, 9, and 10.




--------------------------------------------------------------------------------
DATA TYPE 1:  Color-mapped images.                                             |
_______________________________________________________________________________|
| Offset | Length |                     Description                            |
|--------|--------|------------------------------------------------------------|
|--------|--------|------------------------------------------------------------|
|    0   |     1  |  Number of Characters in Identification Field.             |
|        |        |                                                            |
|        |        |  This field is a one-byte unsigned integer, specifying     |
|        |        |  the length of the Image Identification Field.  Its range  |
|        |        |  is 0 to 255.  A value of 0 means that no Image            |
|        |        |  Identification Field is included.                         |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    1   |     1  |  Color Map Type.                                           |
|        |        |                                                            |
|        |        |  This field contains a binary 1 for Data Type 1 images.    |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    2   |     1  |  Image Type Code.                                          |
|        |        |                                                            |
|        |        |  This field will always contain a binary 1.                |
|        |        |  ( That's what makes it Data Type 1 ).                     |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    3   |     5  |  Color Map Specification.                                  |
|        |        |                                                            |
|    3   |     2  |  Color Map Origin.                                         |
|        |        |  Integer ( lo-hi ) index of first color map entry.         |
|        |        |                                                            |
|    5   |     2  |  Color Map Length.                                         |
|        |        |  Integer ( lo-hi ) count of color map entries.             |
|        |        |                                                            |
|    7   |     1  |  Color Map Entry Size.                                     |
|        |        |  Number of bits in each color map entry.  16 for           |
|        |        |  the Targa 16, 24 for the Targa 24, 32 for the Targa 32.   |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    8   |    10  |  Image Specification.                                      |
|        |        |                                                            |
|    8   |     2  |  X Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) X coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   10   |     2  |  Y Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) Y coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   12   |     2  |  Width of Image.                                           |
|        |        |  Integer ( lo-hi ) width of the image in pixels.           |
|        |        |                                                            |
|   14   |     2  |  Height of Image.                                          |
|        |        |  Integer ( lo-hi ) height of the image in pixels.          |
|        |        |                                                            |
|   16   |     1  |  Image Pixel Size.                                         |
|        |        |  Number of bits in a stored pixel index.                   |
|        |        |                                                            |
|   17   |     1  |  Image Descriptor Byte.                                    |
|        |        |  Bits 3-0 - number of attribute bits associated with each  |
|        |        |             pixel.                                         |
|        |        |  Bit 4    - reserved.  Must be set to 0.                   |
|        |        |  Bit 5    - screen origin bit.                             |
|        |        |             0 = Origin in lower left-hand corner.          |
|        |        |             1 = Origin in upper left-hand corner.          |
|        |        |             Must be 0 for Truevision images.               |
|        |        |  Bits 7-6 - Data storage interleaving flag.                |
|        |        |             00 = non-interleaved.                          |
|        |        |             01 = two-way (even/odd) interleaving.          |
|        |        |             10 = four way interleaving.                    |
|        |        |             11 = reserved.                                 |
|        |        |  This entire byte should be set to 0.  Don't ask me.       |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|   18   | varies |  Image Identification Field.                               |
|        |        |  Contains a free-form identification field of the length   |
|        |        |  specified in byte 1 of the image record.  It's usually    |
|        |        |  omitted ( length in byte 1 = 0 ), but can be up to 255    |
|        |        |  characters.  If more identification information is        |
|        |        |  required, it can be stored after the image data.          |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Color map data.                                           |
|        |        |                                                            |
|        |        |  The offset is determined by the size of the Image         |
|        |        |  Identification Field.  The length is determined by        |
|        |        |  the Color Map Specification, which describes the          |
|        |        |  size of each entry and the number of entries.             |
|        |        |  Each color map entry is 2, 3, or 4 bytes.                 |
|        |        |  Unused bits are assumed to specify attribute bits.        |
|        |        |                                                            |
|        |        |  The 4 byte entry contains 1 byte for blue, 1 byte         |
|        |        |  for green, 1 byte for red, and 1 byte of attribute        |
|        |        |  information, in that order.
|        |        |                                                            |
|        |        |  The 3 byte entry contains 1 byte each of blue, green,     |
|        |        |  and red.                                                  |
|        |        |                                                            |
|        |        |  The 2 byte entry is broken down as follows:               |
|        |        |  ARRRRRGG GGGBBBBB, where each letter represents a bit.    |
|        |        |  But, because of the lo-hi storage order, the first byte   |
|        |        |  coming from the file will actually be GGGBBBBB, and the   |
|        |        |  second will be ARRRRRGG. "A" represents an attribute bit. |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Image Data Field.                                         |
|        |        |                                                            |
|        |        |  This field specifies (width) x (height) color map         |
|        |        |  indices.  Each index is stored as an integral number      |
|        |        |  of bytes (typically 1 or 2).   All fields are unsigned.   |
|        |        |  The low-order byte of a two-byte field is stored first.   |
|        |        |                                                            |
--------------------------------------------------------------------------------






--------------------------------------------------------------------------------
DATA TYPE 2:  Unmapped RGB images.                                             |
_______________________________________________________________________________|
| Offset | Length |                     Description                            |
|--------|--------|------------------------------------------------------------|
|--------|--------|------------------------------------------------------------|
|    0   |     1  |  Number of Characters in Identification Field.             |
|        |        |                                                            |
|        |        |  This field is a one-byte unsigned integer, specifying     |
|        |        |  the length of the Image Identification Field.  Its value  |
|        |        |  is 0 to 255.  A value of 0 means that no Image            |
|        |        |  Identification Field is included.                         |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    1   |     1  |  Color Map Type.                                           |
|        |        |                                                            |
|        |        |  This field contains either 0 or 1.  0 means no color map  |
|        |        |  is included.  1 means a color map is included, but since  |
|        |        |  this is an unmapped image it is usually ignored.  TIPS    |
|        |        |  ( a Targa paint system ) will set the border color        |
|        |        |  the first map color if it is present.                     |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    2   |     1  |  Image Type Code.                                          |
|        |        |                                                            |
|        |        |  This field will always contain a binary 2.                |
|        |        |  ( That's what makes it Data Type 2 ).                     |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    3   |     5  |  Color Map Specification.                                  |
|        |        |                                                            |
|        |        |  Ignored if Color Map Type is 0; otherwise, interpreted    |
|        |        |  as follows:                                               |
|        |        |                                                            |
|    3   |     2  |  Color Map Origin.                                         |
|        |        |  Integer ( lo-hi ) index of first color map entry.         |
|        |        |                                                            |
|    5   |     2  |  Color Map Length.                                         |
|        |        |  Integer ( lo-hi ) count of color map entries.             |
|        |        |                                                            |
|    7   |     1  |  Color Map Entry Size.                                     |
|        |        |  Number of bits in color map entry.  16 for the Targa 16,  |
|        |        |  24 for the Targa 24, 32 for the Targa 32.                 |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    8   |    10  |  Image Specification.                                      |
|        |        |                                                            |
|    8   |     2  |  X Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) X coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   10   |     2  |  Y Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) Y coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   12   |     2  |  Width of Image.                                           |
|        |        |  Integer ( lo-hi ) width of the image in pixels.           |
|        |        |                                                            |
|   14   |     2  |  Height of Image.                                          |
|        |        |  Integer ( lo-hi ) height of the image in pixels.          |
|        |        |                                                            |
|   16   |     1  |  Image Pixel Size.                                         |
|        |        |  Number of bits in a pixel.  This is 16 for Targa 16,      |
|        |        |  24 for Targa 24, and .... well, you get the idea.         |
|        |        |                                                            |
|   17   |     1  |  Image Descriptor Byte.                                    |
|        |        |  Bits 3-0 - number of attribute bits associated with each  |
|        |        |             pixel.  For the Targa 16, this would be 0 or   |
|        |        |             1.  For the Targa 24, it should be 0.  For     |
|        |        |             Targa 32, it should be 8.                      |
|        |        |  Bit 4    - reserved.  Must be set to 0.                   |
|        |        |  Bit 5    - screen origin bit.                             |
|        |        |             0 = Origin in lower left-hand corner.          |
|        |        |             1 = Origin in upper left-hand corner.          |
|        |        |             Must be 0 for Truevision images.               |
|        |        |  Bits 7-6 - Data storage interleaving flag.                |
|        |        |             00 = non-interleaved.                          |
|        |        |             01 = two-way (even/odd) interleaving.          |
|        |        |             10 = four way interleaving.                    |
|        |        |             11 = reserved.                                 |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|   18   | varies |  Image Identification Field.                               |
|        |        |  Contains a free-form identification field of the length   |
|        |        |  specified in byte 1 of the image record.  It's usually    |
|        |        |  omitted ( length in byte 1 = 0 ), but can be up to 255    |
|        |        |  characters.  If more identification information is        |
|        |        |  required, it can be stored after the image data.          |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Color map data.                                           |
|        |        |                                                            |
|        |        |  If the Color Map Type is 0, this field doesn't exist.     |
|        |        |  Otherwise, just read past it to get to the image.         |
|        |        |  The Color Map Specification describes the size of each    |
|        |        |  entry, and the number of entries you'll have to skip.     |
|        |        |  Each color map entry is 2, 3, or 4 bytes.                 |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Image Data Field.                                         |
|        |        |                                                            |
|        |        |  This field specifies (width) x (height) pixels.  Each     |
|        |        |  pixel specifies an RGB color value, which is stored as    |
|        |        |  an integral number of bytes.                              |
|        |        |                                                            |
|        |        |  The 2 byte entry is broken down as follows:               |
|        |        |  ARRRRRGG GGGBBBBB, where each letter represents a bit.    |
|        |        |  But, because of the lo-hi storage order, the first byte   |
|        |        |  coming from the file will actually be GGGBBBBB, and the   |
|        |        |  second will be ARRRRRGG. "A" represents an attribute bit. |
|        |        |                                                            |
|        |        |  The 3 byte entry contains 1 byte each of blue, green,     |
|        |        |  and red.                                                  |
|        |        |                                                            |
|        |        |  The 4 byte entry contains 1 byte each of blue, green,     |
|        |        |  red, and attribute.  For faster speed (because of the     |
|        |        |  hardware of the Targa board itself), Targa 24 images are  |
|        |        |  sometimes stored as Targa 32 images.                      |
|        |        |                                                            |
--------------------------------------------------------------------------------





--------------------------------------------------------------------------------
DATA TYPE 9:  Run Length Encoded, color-mapped images.                         |
_______________________________________________________________________________|
| Offset | Length |                     Description                            |
|--------|--------|------------------------------------------------------------|
|--------|--------|------------------------------------------------------------|
|    0   |     1  |  Number of Characters in Identification Field.             |
|        |        |                                                            |
|        |        |  This field is a one-byte unsigned integer, specifying     |
|        |        |  the length of the Image Identification Field.  Its value  |
|        |        |  is 0 to 255.  A value of 0 means that no Image            |
|        |        |  Identification Field is included.                         |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    1   |     1  |  Color Map Type.                                           |
|        |        |                                                            |
|        |        |  This field is always 1 for color-mapped images.           |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    2   |     1  |  Image Type Code.                                          |
|        |        |                                                            |
|        |        |  A binary 9 for this data type.                            |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    3   |     5  |  Color Map Specification.                                  |
|        |        |                                                            |
|    3   |     2  |  Color Map Origin.                                         |
|        |        |  Integer ( lo-hi ) index of first color map entry.         |
|        |        |                                                            |
|    5   |     2  |  Color Map Length.                                         |
|        |        |  Integer ( lo-hi ) count of color map entries.             |
|        |        |                                                            |
|    7   |     1  |  Color Map Entry Size.                                     |
|        |        |  Number of bits in each color map entry.  16 for the       |
|        |        |  Targa 16, 24 for the Targa 24, 32 for the Targa 32.       |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    8   |    10  |  Image Specification.                                      |
|        |        |                                                            |
|    8   |     2  |  X Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) X coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   10   |     2  |  Y Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) Y coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   12   |     2  |  Width of Image.                                           |
|        |        |  Integer ( lo-hi ) width of the image in pixels.           |
|        |        |                                                            |
|   14   |     2  |  Height of Image.                                          |
|        |        |  Integer ( lo-hi ) height of the image in pixels.          |
|        |        |                                                            |
|   16   |     1  |  Image Pixel Size.                                         |
|        |        |  Number of bits in a pixel.  This is 16 for Targa 16,      |
|        |        |  24 for Targa 24, and .... well, you get the idea.         |
|        |        |                                                            |
|   17   |     1  |  Image Descriptor Byte.                                    |
|        |        |  Bits 3-0 - number of attribute bits associated with each  |
|        |        |             pixel.  For the Targa 16, this would be 0 or   |
|        |        |             1.  For the Targa 24, it should be 0.  For the |
|        |        |             Targa 32, it should be 8.                      |
|        |        |  Bit 4    - reserved.  Must be set to 0.                   |
|        |        |  Bit 5    - screen origin bit.                             |
|        |        |             0 = Origin in lower left-hand corner.          |
|        |        |             1 = Origin in upper left-hand corner.          |
|        |        |             Must be 0 for Truevision images.               |
|        |        |  Bits 7-6 - Data storage interleaving flag.                |
|        |        |             00 = non-interleaved.                          |
|        |        |             01 = two-way (even/odd) interleaving.          |
|        |        |             10 = four way interleaving.                    |
|        |        |             11 = reserved.                                 |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|   18   | varies |  Image Identification Field.                               |
|        |        |  Contains a free-form identification field of the length   |
|        |        |  specified in byte 1 of the image record.  It's usually    |
|        |        |  omitted ( length in byte 1 = 0 ), but can be up to 255    |
|        |        |  characters.  If more identification information is        |
|        |        |  required, it can be stored after the image data.          |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Color map data.                                           |
|        |        |                                                            |
|        |        |  The offset is determined by the size of the Image         |
|        |        |  Identification Field.  The length is determined by        |
|        |        |  the Color Map Specification, which describes the          |
|        |        |  size of each entry and the number of entries.             |
|        |        |  Each color map entry is 2, 3, or 4 bytes.                 |
|        |        |  Unused bits are assumed to specify attribute bits.        |
|        |        |                                                            |
|        |        |  The 4 byte entry contains 1 byte for blue, 1 byte         |
|        |        |  for green, 1 byte for red, and 1 byte of attribute        |
|        |        |  information, in that order.
|        |        |                                                            |
|        |        |  The 3 byte entry contains 1 byte each of blue, green,     |
|        |        |  and red.                                                  |
|        |        |                                                            |
|        |        |  The 2 byte entry is broken down as follows:               |
|        |        |  ARRRRRGG GGGBBBBB, where each letter represents a bit.    |
|        |        |  But, because of the lo-hi storage order, the first byte   |
|        |        |  coming from the file will actually be GGGBBBBB, and the   |
|        |        |  second will be ARRRRRGG. "A" represents an attribute bit. |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Image Data Field.                                         |
|        |        |                                                            |
|        |        |  This field specifies (width) x (height) color map         |
|        |        |  indices.  The indices are stored in packets.  There       |
|        |        |  two types of packets:  Run-length packets, and Raw        |
|        |        |  packets.  Both types of packets consist of a 1-byte       |
|        |        |  header, identifying the type of packet and specifying a   |
|        |        |  count, followed by a variable-length body.                |
|        |        |  The high-order bit of the header is "1" for the           |
|        |        |  run length packet, and "0" for the raw packet.            |
|        |        |                                                            |
|        |        |  For the run-length packet, the header consists of:        |
|        |        |      __________________________________________________    |
|        |        |      | 1 bit |   7 bit repetition count minus 1.      |    |
|        |        |      |   ID  |   Since the maximum value of this      |    |
|        |        |      |       |   field is 127, the largest possible   |    |
|        |        |      |       |   run size would be 128.               |    |
|        |        |      |-------|----------------------------------------|    |
|        |        |      |   1   |  C     C     C     C     C     C    C  |    |
|        |        |      --------------------------------------------------    |
|        |        |                                                            |
|        |        |  For the raw packet, the header consists of:               |
|        |        |      __________________________________________________    |
|        |        |      | 1 bit |   7 bit number of pixels minus 1.      |    |
|        |        |      |   ID  |   Since the maximum value of this      |    |
|        |        |      |       |   field is 127, there can never be     |    |
|        |        |      |       |   more than 128 pixels per packet.     |    |
|        |        |      |-------|----------------------------------------|    |
|        |        |      |   0   |  N     N     N     N     N     N    N  |    |
|        |        |      --------------------------------------------------    |
|        |        |                                                            |
|        |        |  For the run length packet, the header is followed by      |
|        |        |  a single color index, which is assumed to be repeated     |
|        |        |  the number of times specified in the header.  The RLE     |
|        |        |  packet may cross scan lines ( begin on one line and end   |
|        |        |  on the next ).                                            |
|        |        |                                                            |
|        |        |  For the raw packet, the header is followed by the number  |
|        |        |  of color indices specified in the header.  The raw        |
|        |        |  packet may cross scan lines ( begin on one line and end   |
|        |        |  on the next ).                                            |
|        |        |                                                            |
--------------------------------------------------------------------------------





--------------------------------------------------------------------------------
DATA TYPE 10:  Run Length Encoded, RGB images.                                 |
_______________________________________________________________________________|
| Offset | Length |                     Description                            |
|--------|--------|------------------------------------------------------------|
|--------|--------|------------------------------------------------------------|
|    0   |     1  |  Number of Characters in Identification Field.             |
|        |        |                                                            |
|        |        |  This field is a one-byte unsigned integer, specifying     |
|        |        |  the length of the Image Identification Field.  Its range  |
|        |        |  is 0 to 255.  A value of 0 means that no Image            |
|        |        |  Identification Field is included.                         |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    1   |     1  |  Color Map Type.                                           |
|        |        |                                                            |
|        |        |  This field contains either 0 or 1.  0 means no color map  |
|        |        |  is included.  1 means a color map is included, but since  |
|        |        |  this is an unmapped image it is usually ignored.  TIPS    |
|        |        |  ( a Targa paint system ) will set the border color        |
|        |        |  the first map color if it is present.  Wowie zowie.       |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    2   |     1  |  Image Type Code.                                          |
|        |        |                                                            |
|        |        |  Binary 10 for this type of image.                         |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    3   |     5  |  Color Map Specification.                                  |
|        |        |                                                            |
|        |        |  Ignored if Color Map Type is 0; otherwise, interpreted    |
|        |        |  as follows:                                               |
|        |        |                                                            |
|    3   |     2  |  Color Map Origin.                                         |
|        |        |  Integer ( lo-hi ) index of first color map entry.         |
|        |        |                                                            |
|    5   |     2  |  Color Map Length.                                         |
|        |        |  Integer ( lo-hi ) count of color map entries.             |
|        |        |                                                            |
|    7   |     1  |  Color Map Entry Size.                                     |
|        |        |  Number of bits in color map entry.  This value is 16 for  |
|        |        |  the Targa 16, 24 for the Targa 24, 32 for the Targa 32.   |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|    8   |    10  |  Image Specification.                                      |
|        |        |                                                            |
|    8   |     2  |  X Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) X coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   10   |     2  |  Y Origin of Image.                                        |
|        |        |  Integer ( lo-hi ) Y coordinate of the lower left corner   |
|        |        |  of the image.                                             |
|        |        |                                                            |
|   12   |     2  |  Width of Image.                                           |
|        |        |  Integer ( lo-hi ) width of the image in pixels.           |
|        |        |                                                            |
|   14   |     2  |  Height of Image.                                          |
|        |        |  Integer ( lo-hi ) height of the image in pixels.          |
|        |        |                                                            |
|   16   |     1  |  Image Pixel Size.                                         |
|        |        |  Number of bits in a pixel.  This is 16 for Targa 16,      |
|        |        |  24 for Targa 24, and .... well, you get the idea.         |
|        |        |                                                            |
|   17   |     1  |  Image Descriptor Byte.                                    |
|        |        |  Bits 3-0 - number of attribute bits associated with each  |
|        |        |             pixel.  For the Targa 16, this would be 0 or   |
|        |        |             1.  For the Targa 24, it should be 0.  For the |
|        |        |             Targa 32, it should be 8.                      |
|        |        |  Bit 4    - reserved.  Must be set to 0.                   |
|        |        |  Bit 5    - screen origin bit.                             |
|        |        |             0 = Origin in lower left-hand corner.          |
|        |        |             1 = Origin in upper left-hand corner.          |
|        |        |             Must be 0 for Truevision images.               |
|        |        |  Bits 7-6 - Data storage interleaving flag.                |
|        |        |             00 = non-interleaved.                          |
|        |        |             01 = two-way (even/odd) interleaving.          |
|        |        |             10 = four way interleaving.                    |
|        |        |             11 = reserved.                                 |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
|   18   | varies |  Image Identification Field.                               |
|        |        |  Contains a free-form identification field of the length   |
|        |        |  specified in byte 1 of the image record.  It's usually    |
|        |        |  omitted ( length in byte 1 = 0 ), but can be up to 255    |
|        |        |  characters.  If more identification information is        |
|        |        |  required, it can be stored after the image data.          |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Color map data.                                           |
|        |        |                                                            |
|        |        |  If the Color Map Type is 0, this field doesn't exist.     |
|        |        |  Otherwise, just read past it to get to the image.         |
|        |        |  The Color Map Specification, describes the size of each   |
|        |        |  entry, and the number of entries you'll have to skip.     |
|        |        |  Each color map entry is 2, 3, or 4 bytes.                 |
|        |        |                                                            |
|--------|--------|------------------------------------------------------------|
| varies | varies |  Image Data Field.                                         |
|        |        |                                                            |
|        |        |  This field specifies (width) x (height) pixels.  The      |
|        |        |  RGB color information for the pixels is stored in         |
|        |        |  packets.  There are two types of packets:  Run-length     |
|        |        |  encoded packets, and raw packets.  Both have a 1-byte     |
|        |        |  header, identifying the type of packet and specifying a   |
|        |        |  count, followed by a variable-length body.                |
|        |        |  The high-order bit of the header is "1" for the           |
|        |        |  run length packet, and "0" for the raw packet.            |
|        |        |                                                            |
|        |        |  For the run-length packet, the header consists of:        |
|        |        |      __________________________________________________    |
|        |        |      | 1 bit |   7 bit repetition count minus 1.      |    |
|        |        |      |   ID  |   Since the maximum value of this      |    |
|        |        |      |       |   field is 127, the largest possible   |    |
|        |        |      |       |   run size would be 128.               |    |
|        |        |      |-------|----------------------------------------|    |
|        |        |      |   1   |  C     C     C     C     C     C    C  |    |
|        |        |      --------------------------------------------------    |
|        |        |                                                            |
|        |        |  For the raw packet, the header consists of:               |
|        |        |      __________________________________________________    |
|        |        |      | 1 bit |   7 bit number of pixels minus 1.      |    |
|        |        |      |   ID  |   Since the maximum value of this      |    |
|        |        |      |       |   field is 127, there can never be     |    |
|        |        |      |       |   more than 128 pixels per packet.     |    |
|        |        |      |-------|----------------------------------------|    |
|        |        |      |   0   |  N     N     N     N     N     N    N  |    |
|        |        |      --------------------------------------------------    |
|        |        |                                                            |
|        |        |                                                            |
|        |        |  For the run length packet, the header is followed by      |
|        |        |  a single color value, which is assumed to be repeated     |
|        |        |  the number of times specified in the header.  The         |
|        |        |  packet may cross scan lines ( begin on one line and end   |
|        |        |  on the next ).                                            |
|        |        |                                                            |
|        |        |  For the raw packet, the header is followed by             |
|        |        |  the number of color values specified in the header.       |
|        |        |                                                            |
|        |        |  The color entries themselves are two bytes, three bytes,  |
|        |        |  or four bytes ( for Targa 16, 24, and 32 ), and are       |
|        |        |  broken down as follows:                                   |
|        |        |                                                            |
|        |        |  The 2 byte entry -                                        |
|        |        |  ARRRRRGG GGGBBBBB, where each letter represents a bit.    |
|        |        |  But, because of the lo-hi storage order, the first byte   |
|        |        |  coming from the file will actually be GGGBBBBB, and the   |
|        |        |  second will be ARRRRRGG. "A" represents an attribute bit. |
|        |        |                                                            |
|        |        |  The 3 byte entry contains 1 byte each of blue, green,     |
|        |        |  and red.                                                  |
|        |        |                                                            |
|        |        |  The 4 byte entry contains 1 byte each of blue, green,     |
|        |        |  red, and attribute.  For faster speed (because of the     |
|        |        |  hardware of the Targa board itself), Targa 24 image are   |
|        |        |  sometimes stored as Targa 32 images.                      |
|        |        |                                                            |
--------------------------------------------------------------------------------

*/

#include <stdio.h>
#include <stdlib.h>
#include "rtga.h"

int getTGAbyte(FILE *tgafp);
int getTGAint(FILE *tgafp);
int getTGAbuffer(FILE *tgafp, char *buffer, int length);
int getTGAunmappedpixel16(FILE *tgafp, unsigned char *rgba);
int getTGAunmappedpixel24(FILE *tgafp, unsigned char *rgba);
int getTGAunmappedpixel32(FILE *tgafp, unsigned char *rgba);

int getTGAunmappedRLLpacket16(FILE *tgafp, unsigned char *imagedata, int p);
int getTGAunmappedRLLpacket24(FILE *tgafp, unsigned char *imagedata, int p);
int getTGAunmappedRLLpacket32(FILE *tgafp, unsigned char *imagedata, int p);


/*
int main(int argc, char **argv){
	int width;
	int height;
	char *buffer;
	readTGA(argv[1], buffer, &width, &height);
}
*/

int readTGA(char *filename, unsigned char **buffer, int *width, int *height, int *components){
	int idfieldsize;
	int colormapexists;
	int imagetype;
	int colormapstart, colormapsize, colormapbpp;
	int imageposx, imageposy, imagewidth, imageheight, imagebpp, imagedescriptor;
	int compnum, pixels;
	char *idfield;
	char *colormap;
	unsigned char *imagedata;
	unsigned char rgba[4];
	int i, j, p;
	int pixel;
	FILE *tgafp;

	tgafp = fopen(filename, "rb");
	if (tgafp == NULL){
		//printf("Error opening targa file %s\n", filename);
		return(0);
	}

	idfieldsize = getTGAbyte(tgafp);
	colormapexists = getTGAbyte(tgafp);
	imagetype = getTGAbyte(tgafp);
	colormapstart = getTGAint(tgafp);
	colormapsize = getTGAint(tgafp);
	colormapbpp = getTGAbyte(tgafp);
	imageposx = getTGAint(tgafp);
	imageposy = getTGAint(tgafp);
	imagewidth = getTGAint(tgafp);
	imageheight = getTGAint(tgafp);
	imagebpp = getTGAbyte(tgafp);
	imagedescriptor = getTGAbyte(tgafp);

	if (imagebpp == 24) compnum = 3;
	else compnum = 4;

	if (idfieldsize > 0){
		idfield = malloc(idfieldsize);	
		if (idfield == NULL){
			printf("Error allocating identification memory for %s\n", filename);
			exit(-1);
		}	
		getTGAbuffer(tgafp, idfield, idfieldsize);
	}
	else idfield = "";
	
	if (colormapexists){
		colormap = malloc(colormapsize * colormapbpp / 8);
		if (colormap == NULL){
			printf("Error allocating colormap memory for %s\n", filename);
			exit(-1);
		}	
		getTGAbuffer(tgafp, colormap, colormapsize * colormapbpp / 8);
	}

	imagedata = malloc(imagewidth * imageheight * compnum);
	if (imagedata == NULL){
		printf("Error allocating image memory for %s\n", filename);
		exit(-1);
	}	

	#ifdef TGA_DEBUG
	printf("ID Size: %d \nColorMap: %d \nImage Type: %d ", idfieldsize, colormapexists, imagetype); 
	if (imagetype == 1) printf ("(uncompressed color mapped)\n");
	else if (imagetype == 2) printf ("(uncompressed RGB)\n");
	else if (imagetype == 9) printf ("(runlength encoded color mapped)\n");
	else if (imagetype == 10) printf ("(runlength encoded RGB)\n");
	else printf ("(not supported)\n");
	printf("ColorMap Start: %d \nColorMap Size: %d \nColorMap BPP: %d\n", 
		colormapstart, colormapsize, colormapbpp); 
	printf("Image Position: (%d,%d) \nImage Size: %dx%d \nImage BPP: %d \nImage Descriptor: %d\n", 
		imageposx, imageposy, imagewidth, imageheight, imagebpp, imagedescriptor); 
	#endif

	/* uncompressed RGB */
	if (imagetype == 2){
		i = 0;
		j = imageheight - 1;


		while(1){
			pixel = 0;
			p = ((j * imagewidth) + i) * compnum;
			if (imagebpp == 16) pixel = getTGAunmappedpixel16(tgafp, rgba);
			else if (imagebpp == 24) pixel = getTGAunmappedpixel24(tgafp, rgba);
			else if (imagebpp == 32) pixel = getTGAunmappedpixel32(tgafp, rgba);
			if (pixel == 0){  	
				printf("Error reading pixel data from %s\n", filename);
				return(0);
			}

			if (compnum == 3){
				imagedata[p] = rgba[0];
				imagedata[p + 1] = rgba[1];
				imagedata[p + 2] = rgba[2];
			}
			else{
				imagedata[p] = rgba[0];
				imagedata[p + 1] = rgba[1];
				imagedata[p + 2] = rgba[2];
				imagedata[p + 3] = rgba[3];
			}
			
			// #ifdef TGA_DEBUG
			// printf("pixel %d, RGBA: %u,%u,%u,%u\n", i, rgba[0], rgba[1], rgba[2], rgba[3]); 
			// printf("pixel %d, RGBA: %u,%u,%u,%u\n", i, imagedata[i*4], imagedata[i*4+1], 
			//	imagedata[i*4+2], imagedata[i*4+3]); 
			// #endif

			i++;
			if (i == imagewidth && j == 0){
				*width = imagewidth;
				*height = imageheight;
				*buffer = imagedata;
				*components = compnum;
				return(1);
			}
			if (i == imagewidth){
				i = 0;
				j--;
			}				
				
		}
	}

	else if (imagetype == 10){
		i = 0;
		j = 0;

		p = 0;

		/* do this in the correct left-right top-bottom order, then mirror if required */
		while(1){
			pixels = 0;
			p = ((j * imagewidth) + i) * compnum;

			if (imagebpp == 16) pixels = getTGAunmappedRLLpacket16(tgafp, imagedata, p);
			else if (imagebpp == 24) pixels = getTGAunmappedRLLpacket24(tgafp, imagedata, p);
			else if (imagebpp == 32) pixels = getTGAunmappedRLLpacket32(tgafp, imagedata, p);

			#ifdef TGA_DEBUG
			printf("Packet of size %d, buffer at %d, x = %d y =  %d\n", pixels, p, i, j);
			#endif

			if (pixels == 0){  	
				printf("Error reading pixel data from %s\n", filename);
				return(0);
			}
			i += pixels;
	
			if ((i >= imagewidth - 1 && j >= imageheight - 1) || j >= imageheight){
				*width = imagewidth;
				*height = imageheight;
				*buffer = imagedata;
				*components = compnum;
				return(1);
			}
			if (i >= imagewidth - 1){
				j = j + (i  / imagewidth);
				i = i % imagewidth;
			}				
		}
	}
	return(0);
}
	
int getTGAint(FILE *tgafp){
	int value;
	int temp;

	temp = fgetc(tgafp);
	value = fgetc(tgafp);
	value = value << 8;
	value = value + temp;

	if (ferror(tgafp) || feof(tgafp)){
		return(EOF);
	}
	return(value);
}

int getTGAbyte(FILE *tgafp){
	int value;

	value = fgetc(tgafp);
	if (ferror(tgafp) || feof(tgafp)){
		return(EOF);
	}
	return(value);
}

int getTGAbuffer(FILE *tgafp, char *buffer, int length){
	int items;

	if (length > 0){
		items = fread(buffer, length, 1, tgafp);
		if (items == 0 || ferror(tgafp) || feof(tgafp)){
			return(EOF);
		}
	}
	return(1);
}


int getTGAunmappedpixel16(FILE *tgafp, unsigned char *rgba){
	unsigned int b1, b2;
	
	b1 = (unsigned int)fgetc(tgafp);
	b2 = (unsigned int)fgetc(tgafp);
	if (ferror(tgafp) || feof(tgafp)){
		return(0);
	}
	
	rgba[0] = (b2 & 0x7c) >> 2;
	rgba[1] = ((b2 & 0x03) << 3) | ((b1 & 0xe0) >> 5);
	rgba[2] = (b1 & 0x1f);
	rgba[3] = (b2 & 0x80) >> 7;
	return(1);
}

int getTGAunmappedpixel24(FILE *tgafp, unsigned char *rgba){
	rgba[2] = (unsigned char)fgetc(tgafp);
	rgba[1] = (unsigned char)fgetc(tgafp);
	rgba[0] = (unsigned char)fgetc(tgafp);
	rgba[3] = (unsigned char)255;
	if (ferror(tgafp) || feof(tgafp)){
		return(0);
	}
	return(1);
}

int getTGAunmappedpixel32(FILE *tgafp, unsigned char *rgba){
	rgba[2] = (unsigned char)fgetc(tgafp);
	rgba[1] = (unsigned char)fgetc(tgafp);
	rgba[0] = (unsigned char)fgetc(tgafp);
	rgba[3] = (unsigned char)fgetc(tgafp);
	if (ferror(tgafp) || feof(tgafp)){
		return(0);
	}
	return(1);
}


int getTGAunmappedRLLpacket16(FILE *tgafp, unsigned char *imagedata, int p){
	int header;
	int packettype;
	int pixelnum;
	int i, s;
	unsigned char *ptr;
	unsigned char r, g, b;
	unsigned char tempbuffer[128 * 2];
 	
	header = (unsigned int)fgetc(tgafp);
	packettype = (header & 0x80) >> 7;
	pixelnum = (header & 0x7f) + 1;

	/************ REDO THIS *****************/

	//printf("RLL Packet type = %d, pixels = %d offset %d\n", packettype, pixelnum, p);
	
	/* raw packet */
	if (packettype == 0){
		if (getTGAbuffer(tgafp, tempbuffer, pixelnum * 4) == EOF) return(0);			

		/* swap bgra 2 byte format to rgba 4 byte format*/ 
		for (i = 0, s = 0; i < pixelnum; i++, s += 3){
			imagedata[p + s] = tempbuffer[s + 2];
			imagedata[p + s + 1] = tempbuffer[s + 1];
			imagedata[p + s + 2] = tempbuffer[s];
		}
	}

	/* rll packet */
	else{
		b = (unsigned char)fgetc(tgafp);
		g = (unsigned char)fgetc(tgafp);
		r = (unsigned char)fgetc(tgafp);
		ptr = &imagedata[p];
		for (i = 0; i < pixelnum; i++){
			*ptr = r; ptr++;
			*ptr = g; ptr++;
			*ptr = b; ptr++;
		}
	}
	return(pixelnum);
}

int getTGAunmappedRLLpacket24(FILE *tgafp, unsigned char *imagedata, int p){
	int header;
	int packettype;
	int pixelnum;
	int i, s;
	unsigned char *ptr;
	unsigned char r, g, b;
	unsigned char tempbuffer[128 * 3];
 	
	header = (unsigned int)fgetc(tgafp);
	packettype = (header & 0x80) >> 7;
	pixelnum = (header & 0x7f) + 1;

	//printf("RLL Packet type = %d, pixels = %d offset %d\n", packettype, pixelnum, p);
	
	/* raw packet */
	if (packettype == 0){
		if (getTGAbuffer(tgafp, tempbuffer, pixelnum * 4) == EOF) return(0);			

		/* swap bgra to rgba */ 
		for (i = 0, s = 0; i < pixelnum; i++, s += 3){
			imagedata[p + s] = tempbuffer[s + 2];
			imagedata[p + s + 1] = tempbuffer[s + 1];
			imagedata[p + s + 2] = tempbuffer[s];
		}
	}

	/* rll packet */
	else{
		b = (unsigned char)fgetc(tgafp);
		g = (unsigned char)fgetc(tgafp);
		r = (unsigned char)fgetc(tgafp);
		ptr = &imagedata[p];
		for (i = 0; i < pixelnum; i++){
			*ptr = r; ptr++;
			*ptr = g; ptr++;
			*ptr = b; ptr++;
		}
	}
	return(pixelnum);
}

int getTGAunmappedRLLpacket32(FILE *tgafp, unsigned char *imagedata, int p){
	int header;
	int packettype;
	int pixelnum;
	int i, s;
	unsigned char *ptr;
	unsigned char r, g, b, a;
	unsigned char tempbuffer[128 * 4];
 	
	header = (unsigned int)fgetc(tgafp);
	packettype = (header & 0x80) >> 7;
	pixelnum = (header & 0x7f) + 1;

	//printf("RLL Packet type = %d, pixels = %d offset %d\n", packettype, pixelnum, p);
	
	/* raw packet */
	if (packettype == 0){
		if (getTGAbuffer(tgafp, tempbuffer, pixelnum * 4) == EOF) return(0);			

		/* swap bgra to rgba */ 
		for (i = 0, s = 0; i < pixelnum; i++, s += 4){
			imagedata[p + s] = tempbuffer[s + 2];
			imagedata[p + s + 1] = tempbuffer[s + 1];
			imagedata[p + s + 2] = tempbuffer[s];
			imagedata[p + s + 3] = tempbuffer[s + 3];
		}
	}

	/* rll packet */
	else{
		b = (unsigned char)fgetc(tgafp);
		g = (unsigned char)fgetc(tgafp);
		r = (unsigned char)fgetc(tgafp);
		a = (unsigned char)fgetc(tgafp);
		ptr = &imagedata[p];
		for (i = 0; i < pixelnum; i++){
			*ptr = r; ptr++;
			*ptr = g; ptr++;
			*ptr = b; ptr++;
			*ptr = a; ptr++;
		}
	}
	return(pixelnum);
}

